/*
	CDROM.CPP

	Copyright (c) 1994~1995 Marco Savegnago
 
	Gestione DOS riproduzione CDROM AUDIO

	Questo modulo contiene le definizioni delle classi:

	CDROM			gestione completa di un singolo CDROM
	CDROMManager    gestione completa di una serie di CDROM drives


	Data		Autore		Note
	----		------		----
	080395      MS			Collaudato il modulo
	150395		MS			Varie modifiche per gestire eccezioni
							quali richiesta di brani non esistenti

	250595		MS			Aggiunta funzione getCDROM da usare in tutte
							le richieste di puntatori a CDROM fatte
							all'interno del CDROM Manager.
							Questa funzione nel compilatore con la gestione
							delle eccezioni puo' essere il cardine dell'eccezione
	280595		MS			Aggiunto debug su #DEFINE __DEBUG per visualizzare
							lo stato di play
	300595		MS			Aggiunto su #DEFINE __NOPLAY non messa in riproduzione
							dei CDROM per debug
*/


#include "cdrom.h"
#include "reboot.h"

#include <dos.h>
#include <stdio.h>
#include <stdarg.h>
#include <time.h>


/*
	Usa la definizione che segue per evitare che i comandi di PLAY
	da action abbiano effetto
*/

// #define __NOPLAY


/*
	Usa la definizione che segue per avere la visualizzazione a video
	dei tentativi di play
*/

#define __DEBUG

#ifdef	__DEBUG
#include <iostream.h>
#endif

#define TENTATIVICD	 3

/*	Lunghezza (in RED BOOK Len) tagliata ad ogni brano in uscita	*/

const	dword	cutlen	= 160L;

/*	Buffer statico per le richieste di IOCTL	*/

IOCTLHeader CDROM::ioctl;

/*	Buffer statico per le richieste di PLAY		*/

PlayReqHdr	CDROM::playReq;

/*
	CDROM ctor
	Richiede come parametri l'identificatore e il puntatore al devicedriver
*/

CDROM::CDROM(word id_, DeviceHeader far*	devAddr_)
:
	TTriggerItem(TTriggerItem::CDROM)
{
	isInPlay		= False;

	id 				= id_;
	devAddr 		= devAddr_;

	/*	Alloca i far buffer per eseguire le richieste al device driver */

	diskInfo 		= new  	DiskInfoRec;
	tnoInfo			= new  	TNOInfoRec;
	qChanInfoRec	= new  	QChanInfoRec;
	deviceStatus	= new   DeviceStatusRec;

	ejectDiskRec		= new EjectDiskRec;		// Buffer per richiedere l'eject del disco
	closeTrayRec		= new CloseTrayRec;		// Buffer per richiedere la chiusura del cassetto
	resetDiskRec 		= new ResetDiskRec;		// Buffer per richiedere il reset del drive
	lockUnlockDoorRec 	= new LockUnlockDoorRec;  // Buffer per richiedere il blocco e lo sblocco del cassetto
	audioChannelInfoRec = new AudioChannelInfoRec;	// Buffer per richiedere le audio channel info

	/*
		Alloca lo spazio per le collezioni di tracce ...
		Preventivamente 10 + 1 per ogni salto
	*/

	tracks			= new  TBNSCollection(10,10);
}

CDROM::~CDROM()
{
	/*	Cancella i far buffer allocati precedentemente	*/

	delete 	diskInfo;
	delete 	tnoInfo;
	delete 	qChanInfoRec;
	delete	deviceStatus;
	delete	ejectDiskRec;
	delete	closeTrayRec;
	delete	resetDiskRec;
	delete	lockUnlockDoorRec;
	delete 	audioChannelInfoRec;

	/*	Cancella le singole traccie memorizzate	*/

	if(tracks!=NULL)
	{
		for(int i=0; i < tracks->getCount(); i++)
		{
			TNOInfoRec* tnoInfo = (TNOInfoRec*)tracks->at(i);

			if(tnoInfo)
				delete tnoInfo;
		}
	}

	delete	tracks;

}

/* red2hsg() -
**
** DESCRIPTION
**	Converts a binary red book address to high sierra addressing
**	The msb of the red book is always zero, the next less significant
**	byte is the minute (0-59+), then second (0-59) and lsb is the
**	frame (0-75). The conversion is
**		hsg = min * 60 * 75 + sec * 75 + frame;
*/
dword CDROM::red2hsg(dword l)
{
	return((dword) (HIWORD(l) & 0xff) * 60 * 75 + (dword) (LOWORD(l) >> 8) * 75 + (dword) (LOWORD(l) & 0xff));
}

/*
**
** DESCRIPTION
**	Prints red book address in MM:SS:FF Min/Sec/Frame format
*/

void CDROM::printAddress(char* str, dword	addr)
{
	sprintf(str, "%2d:%02d.%02d", HIWORD(addr) & 0xff, 	LOWORD(addr) >> 8, LOWORD(addr) & 0xff);
}

/*
	Ritorna il punto di inizio di riproduzione di una traccia specifica
*/

dword	CDROM::getTrackStartREDAddress(word track)
{
	TNOInfoRec* tnoInfo = (TNOInfoRec*)tracks->at(track);

	return	tnoInfo->startAddr;
}

/*
	Ritorna la lunghezza di riproduzione di una traccia specifica
*/

dword	CDROM::getTrackLenRED(word track)
{
	dword len;

	/*
		Se ho richiesto la lunghezza di una traccia inferiore all'ultima
		allora fai la differenza tra il punto di inizio di quella successiva
		e la richiesta
	*/

	if(track < (highTrack() - 1))
	{
		TNOInfoRec* tnoInfo = (TNOInfoRec*)tracks->at(track + 1);

		len = red2hsg(tnoInfo->startAddr) - red2hsg(getTrackStartREDAddress(track));
	}

	/*
		Altrimenti fai la differenza tra il punto di fine del disco
		(contenuto nel buffer diskinfo) e l'inizio della prima traccia
	*/
	else
		len = red2hsg(diskInfo->leadOut) - red2hsg(getTrackStartREDAddress(track));

	return len;

}

/*
	Ritorna la posizione corrispondente all'inizio del disco
*/
dword	CDROM::getBeginOfDisk()
{
	if(tracks->getCount())
	{
		TNOInfoRec* tnoInfo = (TNOInfoRec*)tracks->at(0);

		return	tnoInfo->startAddr;
	}

	return 0;

}

/*
	Ritorna la lunghezza complessiva del disco facendo la differenza tra il
	punto di lead out e l'inizio della prima traccia MS150395
*/

dword	CDROM::getLenOfDisk()
{
	return red2hsg(diskInfo->leadOut) - red2hsg(getTrackStartREDAddress(0));
}

Boolean	CDROM::readTOC()
{
	if(!IOCTLInputCmdAudioDiskInfo())
		return False;

		//  Poi se il numero di brani contenuti e' != 0 allora
		//	per il numero di brani richiedi le info

	if(diskInfo->highTNO > diskInfo->lowTNO)
	{
		for(int j = lowTrack(); j <= highTrack(); j++)
		{

			//	Inserisci il numero della traccia di cui richiedere l'info

			tnoInfo->tno=j;

			//	Richiedi le info e salvale nel buffer

			if(!IOCTLInputCmdAudioTrackInfo())
				return False;

			//  Se la lettura e' andata a buon fine allora copiale nel
			//	track info item che andra' dentro la collezione

			TNOInfoRec* TnoInfo = new TNOInfoRec(*tnoInfo);

			tracks->insert(TnoInfo);
		}
	}

	return True;
}

/*
	Ferma la riproduzione del CDROM selezionato dato il suo puntatore (interna)
*/

Boolean CDROM::stop()
{
	playReq.plRqh.rqhLen 	= sizeof (PlayReqHdr);
	playReq.plRqh.rqhUnit	= id;
	playReq.plRqh.rqhCmd 	= DEVSTOP;
	playReq.plRqh.rqhStatus = 0;
	playReq.plAddrmd		= ADDR_RED;
	playReq.plStart			= 0;
	playReq.plNum			= 0;

	sendReq(&playReq);

	if(!IOCTLInputCmdAudioDiskInfo())
			return False;

	return True;
}

/*
	Esegue un play sul CDROM selezionato da 'start' per 'len'
*/

Boolean CDROM::play(dword start, dword len)
{
	playReq.plRqh.rqhLen 	= sizeof (PlayReqHdr);
	playReq.plRqh.rqhUnit	= id;
	playReq.plRqh.rqhCmd 	= DEVPLAY;
	playReq.plRqh.rqhStatus = 0;
	playReq.plAddrmd		= ADDR_RED;
	playReq.plStart			= start;
	playReq.plNum			= len;

	sendReq(&playReq);

	if(!IOCTLInputCmdAudioDiskInfo())
			return False;

	return isInPlay;
}

/*
	Manda in riproduzione la traccia del singolo CD

	I riferimenti agli indici sono sempre zero based
*/

Boolean CDROM::play(word track)
{
	dword	start = getTrackStartREDAddress(track-1);

	dword	len   = getTrackLenRED(track-1);

	len	-= cutlen;

	return play(start, len);
}

/*
	Manda in riproduzione completa (dall'inizio alla fine) il singolo CDROM

	I riferimenti agli indici sono sempre zero based
*/

Boolean CDROM::play()
{
	dword	start = getBeginOfDisk();

	dword	len   = getLenOfDisk();

	return play(start, len);
}

/*
	Esegue una chiamata IOCTL diretta al device driver per conoscere
	il diskinfo del CDROM selezionato
*/

Boolean CDROM::IOCTLInputCmdAudioDiskInfo()
{
	ioctl.ioctlRqh.rqhLen 		= sizeof (IOCTLHeader);
	ioctl.ioctlRqh.rqhUnit 		= id;
	ioctl.ioctlRqh.rqhCmd		= DEVRDIOCTL;
	ioctl.ioctlRqh.rqhStatus	= 0;
	ioctl.ioctlMedia			= 0;
	ioctl.ioctlSector			= 0;
	ioctl.ioctlVolId			= 0;

	ioctl.ioctlXfer				= (byte far *)diskInfo;
	ioctl.ioctlNbytes			= sizeof(DiskInfoRec);

	sendReq(&ioctl);

	isInPlay = Boolean(ioctl.ioctlRqh.rqhStatusDetails.busy);

	return True;
}

/*
	Esegue una chiamata IOCTL diretta al device driver per conoscere
	il deviceStatus del CDROM selezionato
*/

Boolean CDROM::IOCTLInputCmdDeviceStatus()
{
	ioctl.ioctlRqh.rqhLen 		= sizeof (IOCTLHeader);
	ioctl.ioctlRqh.rqhUnit 		= id;
	ioctl.ioctlRqh.rqhCmd		= DEVRDIOCTL;
	ioctl.ioctlRqh.rqhStatus	= 0;
	ioctl.ioctlMedia			= 0;
	ioctl.ioctlSector			= 0;
	ioctl.ioctlVolId			= 0;

	ioctl.ioctlXfer				= (byte far *)deviceStatus;
	ioctl.ioctlNbytes			= sizeof(DeviceStatusRec);

	sendReq(&ioctl);

	isInPlay 	= Boolean(ioctl.ioctlRqh.rqhStatusDetails.busy);

	return True;
}

/*
	Esegue una chiamata IOCTL diretta al device driver per conoscere
	le informazioni sul modo in cui sono registrate le traccie audio
*/

Boolean CDROM::IOCTLInputCmdAudioChannelInfo()
{
	ioctl.ioctlRqh.rqhLen 		= sizeof (IOCTLHeader);
	ioctl.ioctlRqh.rqhUnit 		= id;
	ioctl.ioctlRqh.rqhCmd		= DEVRDIOCTL;
	ioctl.ioctlRqh.rqhStatus	= 0;
	ioctl.ioctlMedia			= 0;
	ioctl.ioctlSector			= 0;
	ioctl.ioctlVolId			= 0;

	audioChannelInfoRec->cmdCode= IOI_audio_info;
	ioctl.ioctlXfer				= (byte far *)audioChannelInfoRec;
	ioctl.ioctlNbytes			= sizeof(AudioChannelInfoRec);

	sendReq(&ioctl);

	isInPlay 	= Boolean(ioctl.ioctlRqh.rqhStatusDetails.busy);
	return True;
}

/*
	Esegue una chiamata IOCTL diretta al device driver per conoscere
	il trackinfo del CDROM selezionato
*/

Boolean CDROM::IOCTLInputCmdAudioTrackInfo()
{
	ioctl.ioctlRqh.rqhLen 		= sizeof (IOCTLHeader);
	ioctl.ioctlRqh.rqhUnit 		= id;
	ioctl.ioctlRqh.rqhCmd		= DEVRDIOCTL;
	ioctl.ioctlRqh.rqhStatus	= 0;
	ioctl.ioctlMedia			= 0;
	ioctl.ioctlSector			= 0;
	ioctl.ioctlVolId			= 0;

	ioctl.ioctlXfer				= (byte far *)tnoInfo;
	ioctl.ioctlNbytes			= sizeof(TNOInfoRec);

	sendReq(&ioctl);

	isInPlay = Boolean(ioctl.ioctlRqh.rqhStatusDetails.busy);

	return True;
}

/*
	Esegue una chiamata IOCTL diretta al device driver per conoscere
*/

Boolean CDROM::IOCTLOutputCmdEjectDisk()
{
	ioctl.ioctlRqh.rqhLen 		= sizeof (IOCTLHeader);
	ioctl.ioctlRqh.rqhUnit 		= id;
	ioctl.ioctlRqh.rqhCmd		= DEVWRIOCTL;
	ioctl.ioctlRqh.rqhStatus	= 0;
	ioctl.ioctlMedia			= 0;
	ioctl.ioctlSector			= 0;
	ioctl.ioctlVolId			= 0;

	ioctl.ioctlXfer				= (byte far *)ejectDiskRec;
	ioctl.ioctlNbytes			= sizeof(EjectDiskRec);

	sendReq(&ioctl);

	isInPlay = Boolean(ioctl.ioctlRqh.rqhStatusDetails.busy);

	return True;
}

Boolean CDROM::IOCTLOutputCmdCloseTray()
{
	ioctl.ioctlRqh.rqhLen 		= sizeof (IOCTLHeader);
	ioctl.ioctlRqh.rqhUnit 		= id;
	ioctl.ioctlRqh.rqhCmd		= DEVWRIOCTL;
	ioctl.ioctlRqh.rqhStatus	= 0;
	ioctl.ioctlMedia			= 0;
	ioctl.ioctlSector			= 0;
	ioctl.ioctlVolId			= 0;

	ioctl.ioctlXfer				= (byte far *)closeTrayRec;
	ioctl.ioctlNbytes			= sizeof(CloseTrayRec);

	sendReq(&ioctl);

	isInPlay = Boolean(ioctl.ioctlRqh.rqhStatusDetails.busy);

	return True;
}

Boolean CDROM::IOCTLOutputCmdReset()
{
	ioctl.ioctlRqh.rqhLen 		= sizeof (IOCTLHeader);
	ioctl.ioctlRqh.rqhUnit 		= id;
	ioctl.ioctlRqh.rqhCmd		= DEVWRIOCTL;
	ioctl.ioctlRqh.rqhStatus	= 0;
	ioctl.ioctlMedia			= 0;
	ioctl.ioctlSector			= 0;
	ioctl.ioctlVolId			= 0;

	ioctl.ioctlXfer				= (byte far *)resetDiskRec;
	ioctl.ioctlNbytes			= sizeof(ResetDiskRec);

	sendReq(&ioctl);

	isInPlay = Boolean(ioctl.ioctlRqh.rqhStatusDetails.busy);

	return True;
}

Boolean CDROM::IOCTLOutputLockDoor(Boolean lock)
{
	ioctl.ioctlRqh.rqhLen 		= sizeof (IOCTLHeader);
	ioctl.ioctlRqh.rqhUnit 		= id;
	ioctl.ioctlRqh.rqhCmd		= DEVWRIOCTL;
	ioctl.ioctlRqh.rqhStatus	= 0;
	ioctl.ioctlMedia			= 0;
	ioctl.ioctlSector			= 0;
	ioctl.ioctlVolId			= 0;

	lockUnlockDoorRec->action	= lock;
	ioctl.ioctlXfer				= (byte far *)lockUnlockDoorRec;
	ioctl.ioctlNbytes			= sizeof(LockUnlockDoorRec);

	sendReq(&ioctl);

	isInPlay = Boolean(ioctl.ioctlRqh.rqhStatusDetails.busy);

	return True;
}

/*
	Esegue una chiamata IOCTL diretta al device driver per conoscere
	le informazioni sul modo in cui sono registrate le traccie audio
*/

Boolean CDROM::IOCTLOutputCmdAudioChannelInfo()
{
	ioctl.ioctlRqh.rqhLen 		= sizeof (IOCTLHeader);
	ioctl.ioctlRqh.rqhUnit 		= id;
	ioctl.ioctlRqh.rqhCmd		= DEVWRIOCTL;
	ioctl.ioctlRqh.rqhStatus	= 0;
	ioctl.ioctlMedia			= 0;
	ioctl.ioctlSector			= 0;
	ioctl.ioctlVolId			= 0;

	audioChannelInfoRec->cmdCode= IOO_set_audio_param;
	ioctl.ioctlXfer				= (byte far *)audioChannelInfoRec;
	ioctl.ioctlNbytes			= sizeof(AudioChannelInfoRec);

	sendReq(&ioctl);

	isInPlay 	= Boolean(ioctl.ioctlRqh.rqhStatusDetails.busy);

	return True;
}

/*
	Esegue la chiamata con la richiesta del servizio al device driver
*/

void	CDROM::sendReq(void far * buffer)
{
	//	Typedef per non appesantire i typedefs

	typedef 	void (far *fnpt)();

	//	Puntatore alla strategy routine e alla init routine del driver

	fnpt	strategy;
	fnpt	init;

	/*	L'indirizzo della strategy e init routine del driver e' formato da:

		-	il segmento del puntatore al device driver
		-	l'offset del puntatore all'indirizzo delle strategy e init
			routines

		Questo perche' il segmento assoluto del driver non e' ripetuto
		nel puntatore alle due funzioni	(strategy & init)
	*/

	strategy	= fnpt (MK_FP(FP_SEG(devAddr), FP_OFF(devAddr->sdevstrat)));
	init		= fnpt (MK_FP(FP_SEG(devAddr), FP_OFF(devAddr->sdevint)));

	/*
		Copia i parametri in ES:BX
	*/

	_ES = FP_SEG(buffer);
	_BX = FP_OFF(buffer);

	/*
		Chiama la strategy e poi la init routine
	*/

	strategy();
	init();
}

/*
	Vettore statico di device driver
*/
DeviceList 	CDROMManager::DeviceTable[MAXCDROMDEVICE];


/*
	CDROMManager ctor
*/

CDROMManager::CDROMManager()
:
	TAction(TAction::CDROMManager)
{
	NumOfDrives = 0;
	FirstDriveLetter = 0;

	/*	Alloca lo spazio per il container di cdrom */

	cdroms = new TBNSCollection(MAXCDROMDEVICE,2);
}

/*
	CDROMManager dtor
*/

CDROMManager::~CDROMManager()
{
	// cdroms->shutDown();

	if(cdroms!=NULL)
	{
		for(int i=0; i < cdroms->getCount(); i++)
		{
			CDROM* cdrom= getCDROM(i);

			if(cdrom)
				delete cdrom;
		}
	}

	delete cdroms;
}


void CDROMManager::freeCDROMS()
{
	cdroms->shutDown();
}

/* FindCDRoms() -
**
**	Using INT 2Fh with AH=15h (the MSCDEX function request interface)
**	we can ask MSCDEX for the number and location of all CDROM
**	device drivers on the system.
**
**	Unfortunately at present, if MSCDEX is not present, the carry
**	flag is not set when an INT 2Fh is issued with AH=15h...I'll
**	figure out what's going on and figure out a good way to tell
**	if MSCDEX is present or not.
*/

Boolean	CDROMManager::FindCDRomDrives()
{
	union 		REGS		inregs;
	union 		REGS		outregs;
	struct 		SREGS		segregs;

	inregs.x.ax = 0x150C;	/* Check MSCDEX version					*/
	int86(0x2f, &inregs, &outregs);

	MSCDEXMinorVersion	=	outregs.h.bl;
	MSCDEXMajorVersion	=	outregs.h.bh;

	inregs.x.ax = 0x1500;	/* Get Number of CDROM drive letters	*/
	inregs.x.bx = 0;		/* Init to zero							*/

	int86(0x2f, &inregs, &outregs);

	/* If number of drives returned is still 0, then MSCDEX
	** is not installed.
	*/

	if (outregs.x.bx == 0)
		return False;

	NumOfDrives			= outregs.x.bx;
	FirstDriveLetter	= outregs.x.cx;

	/* Get CDROM drive letter device list	*/

	inregs.x.ax = 0x1501;
	inregs.x.bx = FP_OFF(DeviceTable);
	segregs.es 	= FP_SEG(DeviceTable);

	int86x(0x2f, &inregs, &outregs, &segregs);

	/* Check if carry set		*/

	return Boolean(outregs.x.cflag);
}

/*
	Legge la lista dei brani con le relative posizioni contenuti nel CD
*/

Boolean CDROMManager::ReadCDRomsTOC()
{
	// Per tutti i CD ...
	for(int i=0;i<NumOfDrives;i++)
	{
		CDROM* cdrom = getCDROM(i);

		//	Leggi quanti brani ci sono e da dove iniziano

		if(cdrom)
			cdrom->readTOC();
	}

	return True;
}

/*
	Legge le informazioni riguardanti il CDROM selezionato
*/

Boolean CDROMManager::ReadCDROMS()
{
	/*	Alloca i cdrom	*/

	for(int i=0;i<NumOfDrives;i++)
		cdroms->insert(new CDROM(DeviceTable[i].subUnit, DeviceTable[i].devAddr));

	/*	Leggi le informazioni relative alle traccie registrate nei CD */

	return ReadCDRomsTOC();
}

/*
	Ferma la riproduzione del CDROM selezionato dato il suo indice

	I riferimenti agli indici sono sempre zero based
*/

Boolean CDROMManager::StopCD(word cd)
{
	CDROM* cdrom = getCDROM(cd-1);

	if(cdrom)
		return cdrom->stop();

	return	False;
}

/*
	Resetta il drive
*/

void CDROMManager::ResetCD(word cd)
{
	CDROM* cdrom = getCDROM(cd-1);

	if(cdrom)
		cdrom->resetDrive();
}

/*
	Ferma la riproduzione di tutti i CDROM
*/

void	CDROMManager::StopAll()
{
	for(int i=0; i < cdroms->getCount(); i++)
	{
		CDROM* cdrom = getCDROM(i);

		if(cdrom)
			cdrom->stop();
	}
}

/*
	Manda in riproduzione di tutti i CDROM
*/

void	CDROMManager::PlayAll()
{
	for(int i=1; i <= cdroms->getCount(); i++)
		PlayCD(i);
}

/*
	Manda in riproduzione completa (dall'inizio alla fine) il singolo CDROM

	I riferimenti agli indici sono sempre zero based
*/

Boolean CDROMManager::PlayCD(word cd)
{
	CDROM* cdrom = getCDROM(cd-1);

	if(cdrom!=NULL && cdrom->getTrackCount())
		return cdrom->play();

	return False;
}

/*
	Manda in riproduzione la traccia del singolo CD

	I riferimenti agli indici sono sempre zero based
*/

Boolean CDROMManager::PlayCD(word cd, word track)
{
	CDROM* cdrom = getCDROM(cd-1);

	if(cdrom)
		return cdrom->play(track);

	return False;
}

/*
	Ferma la riproduzione di tutti i CDROM
*/

void	CDROMManager::checkStatus()
{
	for(int i=0; i < cdroms->getCount(); i++)
	{
		CDROM* cdrom = getCDROM(i);

		if(cdrom)
			cdrom->checkStatus();
	}
}

/*
	Espelle tutti i cd
*/

void	CDROMManager::EjectAll()
{
	for(int i=0; i < cdroms->getCount(); i++)
	{
		CDROM* cdrom = getCDROM(i);

		if(cdrom)
			cdrom->eject();
	}
}

/*
	Espelle il CD specificato
*/

void	CDROMManager::EjectCD(word cd)
{
	CDROM* cdrom = getCDROM(cd-1);

	if(cdrom)
		cdrom->eject();
}

void	CDROMManager::CloseTrayAll()
{
	for(int i=0; i < cdroms->getCount(); i++)
	{
		CDROM* cdrom = getCDROM(i);

		if(cdrom)
			cdrom->closeTray();
	}
}

void	CDROMManager::LockAll()
{
	for(int i=0; i < cdroms->getCount(); i++)
	{
		CDROM* cdrom = getCDROM(i);

		if(cdrom)
			cdrom->lockDoor();
	}
}

void	CDROMManager::UnlockAll()
{
	for(int i=0; i < cdroms->getCount(); i++)
	{
		CDROM* cdrom = getCDROM(i);

		if(cdrom)
			cdrom->unlockDoor();
	}
}

void	CDROMManager::CloseTrayCD(word cd)
{
	CDROM* cdrom = getCDROM(cd-1);

	if(cdrom)
		cdrom->closeTray();
}

byte	CDROMManager::getVolume(word cd, byte channel)
{
	CDROM* cdrom = getCDROM(cd-1);

	if(cdrom)
		return cdrom->getVolume(channel);

	return byte(0);
}

Boolean	CDROMManager::setVolume(word cd, byte channel, byte vol)
{
	CDROM* cdrom = getCDROM(cd-1);

	if(cdrom)
		return cdrom->setVolume(channel, vol);

	return False;
}

void	CDROMManager::getCDStart(char* str, word cd)
{
	CDROM* cdrom = getCDROM(cd-1);

	if(cdrom)
		cdrom->printAddress(str, cdrom->getBeginOfDisk());
}

void	CDROMManager::getTrackStart(char* str, word cd, word track)
{
	CDROM* cdrom = getCDROM(cd-1);

	if(cdrom)
		cdrom->printAddress(str, cdrom->getTrackStartREDAddress(track-1));
}

void	CDROMManager::getCDLen(char* str, word cd)
{
	CDROM* cdrom = getCDROM(cd-1);

	if(cdrom)
		cdrom->printAddress(str, cdrom->getLenOfDisk());
}

void	CDROMManager::getTrackLen(char* str, word cd, word track)
{
	CDROM* cdrom = getCDROM(cd-1);

	if(cdrom)
		cdrom->printAddress(str, cdrom->getTrackLenRED(track-1));
}

Boolean	CDROMManager::action(word pos, Boolean set, int	value ...)
{
	Boolean rc;

	va_list argptr;
	va_start(argptr, value);

	int volume = va_arg(argptr,int);

	va_end(argptr);


	if(set)
	{

#ifndef	__NOPLAY
		StopCD(pos);

		if(volume)
			setVolume(pos, volume);
#endif


		if(value)
		{
			for(int n=0;n<TENTATIVICD;n++)
			{

#ifndef	__NOPLAY
				rc = PlayCD(pos, value);
#else
				rc = True;
#endif

				cout	<< "PLAY CD " << pos << " BRANO " << value;

				if(rc)
				{
					cout	<< " OK"	<< endl;
					break;
				}
				else
				{
					if(n == (TENTATIVICD - 1))
					{
						FILE *stream;

						time_t t;

						time(&t);

						stream = fopen("ERROR.SEQ", "a+");

						fprintf(stream, "%s: CD ERROR IN PLAY %d, %d\n", ctime(&t), pos, value);

						fclose(stream);

						reboot(True);
					}
					else
						cout	<< " ERROR " << n << endl;
				}
			}
		}

	}
	else
	{
#ifndef	__NOPLAY
		if(volume)
			setVolume(pos, volume);

		rc = StopCD(pos);

		ResetCD(pos);
#endif
	}

	return rc;
}

void	CDROMManager::ResetDrives()
{
	for(int i=0; i < cdroms->getCount(); i++)
	{
		CDROM* cdrom = getCDROM(i);

		if(cdrom)
			cdrom->resetDrive();
	}
}

void	CDROMManager::setAllVolume(byte volume)
{
	for(int i=0; i < cdroms->getCount(); i++)
	{
		CDROM* cdrom = getCDROM(i);

		if(cdrom)
		{
			for(int i = 0; i < 4; i++)
				cdrom->setVolume(i, volume);
		}
	}
}

void	CDROMManager::setVolume(word cd, byte volume)
{
	CDROM* cdrom = getCDROM(cd-1);

	if(cdrom)
	{
		for(int i = 0; i < 4; i++)
			cdrom->setVolume(i, volume);
	}
}

#if 0

main()
{
	CDROMManager cdrom;

	int cd = cdrom.FindCDRomDrives();

	cdrom.ReadCDROMS();

	cdrom.StopAll();

	cdrom.PlayCD(1,8);

	cdrom.PlayCD(1,2);

	cdrom.PlayCD(2,9);

	return 0;
}

#endif
